'''
Module of Windows API for plyer.cpu.
'''

from ctypes import (
    c_ulonglong, c_ulong, byref,
    Structure, POINTER, Union, windll, create_string_buffer,
    sizeof, cast, c_void_p, c_uint32
)
from ctypes.wintypes import (
    BYTE, DWORD, WORD
)

from plyer.facades import CPU


KERNEL = windll.kernel32
ERROR_INSUFFICIENT_BUFFER = 0x0000007A


class CacheType(object):
    # pylint: disable=too-few-public-methods
    '''
    Win API PROCESSOR_CACHE_TYPE enum.
    '''

    unified = 0
    instruction = 1
    data = 2
    trace = 3


class RelationshipType(object):
    # pylint: disable=too-few-public-methods
    '''
    Win API LOGICAL_PROCESSOR_RELATIONSHIP enum.
    '''

    processor_core = 0     # logical proc sharing single core
    numa_node = 1          # logical proc sharing single NUMA node
    cache = 2              # logical proc sharing cache
    processor_package = 3  # logical proc sharing physical package
    group = 4              # logical proc sharing processor group
    all = 0xffff           # logical proc info for all groups


class CacheDescriptor(Structure):
    # pylint: disable=too-few-public-methods
    '''
    Win API CACHE_DESCRIPTOR struct.
    '''

    _fields_ = [
        ('Level', BYTE),
        ('Associativity', BYTE),
        ('LineSize', WORD),
        ('Size', DWORD),
        ('Type', DWORD)
    ]


class ProcessorCore(Structure):
    # pylint: disable=too-few-public-methods
    '''
    Win API ProcessorCore struct.
    '''

    _fields_ = [('Flags', BYTE)]


class NumaNode(Structure):
    # pylint: disable=too-few-public-methods
    '''
    Win API NumaNode struct.
    '''

    _fields_ = [('NodeNumber', DWORD)]


class SystemLPIUnion(Union):
    # pylint: disable=too-few-public-methods
    '''
    Win API SYSTEM_LOGICAL_PROCESSOR_INFORMATION union without name.
    '''

    _fields_ = [
        ('ProcessorCore', ProcessorCore),
        ('NumaNode', NumaNode),
        ('Cache', CacheDescriptor),
        ('Reserved', c_ulonglong)
    ]


class SystemLPI(Structure):
    # pylint: disable=too-few-public-methods
    '''
    Win API SYSTEM_LOGICAL_PROCESSOR_INFORMATION struct.
    '''

    _fields_ = [
        ('ProcessorMask', c_ulong),
        ('Relationship', c_ulong),
        ('LPI', SystemLPIUnion)
    ]


class WinCPU(CPU):
    '''
    Implementation of Windows CPU API.
    '''

    @staticmethod
    def _countbits(mask):
        # make sure the correct ULONG_PTR size is used on 64bit
        # https://docs.microsoft.com/en-us/windows/
        # desktop/WinProg/windows-data-types
        # note: not a pointer per-se, != PULONG_PTR
        ulong_ptr = c_ulonglong if sizeof(c_void_p) == 8 else c_ulong
        # note: c_ulonglong only on 64bit, otherwise c_ulong

        # DWORD == c_uint32
        # https://docs.microsoft.com/en-us/windows/
        # desktop/WinProg/windows-data-types
        lshift = c_uint32(sizeof(ulong_ptr) * 8 - 1)
        assert lshift.value in (31, 63), lshift  # 32 or 64 bits - 1

        lshift = lshift.value
        test = 1 << lshift
        assert test % 2 == 0, test

        count = 0
        i = 0
        while i <= lshift:
            i += 1

            # do NOT remove!!!
            # test value has to be %2 == 0,
            # except the last case where the value is 1,
            # so that int(test) == int(float(test))
            # and the mask bit is counted correctly
            assert test % 2 == 0 or float(test) == 1.0, test

            # https://stackoverflow.com/a/1746642/5994041
            # note: useful to print(str(bin(int(...)))[2:])
            count += 1 if (mask & int(test)) else 0
            test /= 2

        return count

    def _logprocinfo(self, relationship):
        get_logical_process_info = KERNEL.GetLogicalProcessorInformation

        # first call with no structure to get the real size of the required
        buff_length = c_ulong(0)
        result = get_logical_process_info(None, byref(buff_length))
        assert not result, result
        error = KERNEL.GetLastError()
        assert error == ERROR_INSUFFICIENT_BUFFER, error
        assert buff_length, buff_length

        # create buffer from the real winapi buffer length
        buff = create_string_buffer(buff_length.value)

        # call again with buffer pointer + the same length as arguments
        result = get_logical_process_info(buff, byref(buff_length))
        assert result, (result, KERNEL.GetLastError())

        # memory size of one LPI struct in the array of LPI structs
        offset = sizeof(SystemLPI)  # ok
        values = {
            key: 0 for key in (
                'relationship', 'mask',
                'L1', 'L2', 'L3'
            )
        }

        for i in range(0, buff_length.value, offset):
            slpi = cast(
                buff[i: i + offset],
                POINTER(SystemLPI)
            ).contents

            if slpi.Relationship != relationship:
                continue

            values['relationship'] += 1
            values['mask'] += self._countbits(slpi.ProcessorMask)

            if slpi.LPI.Cache.Level == 1:
                values['L1'] += 1
            elif slpi.LPI.Cache.Level == 2:
                values['L2'] += 1
            elif slpi.LPI.Cache.Level == 3:
                values['L3'] += 1

        return values

    def _sockets(self):
        # physical CPU sockets (or slots) on motherboard
        return self._logprocinfo(
            RelationshipType.processor_package
        )['relationship']

    def _physical(self):
        # cores
        return self._logprocinfo(
            RelationshipType.processor_core
        )['relationship']

    def _logical(self):
        # cores * threads
        # if hyperthreaded core -> more than one logical processor
        return self._logprocinfo(
            RelationshipType.processor_core
        )['mask']

    def _cache(self):
        # L1, L2, L3 cache count
        result = self._logprocinfo(
            RelationshipType.cache
        )
        return {
            key: result[key]
            for key in result
            if key in ('L1', 'L2', 'L3')
        }

    def _numa(self):
        # numa nodes
        return self._logprocinfo(
            RelationshipType.numa_node
        )['relationship']


def instance():
    '''
    Instance for facade proxy.
    '''
    return WinCPU()


# Resources:
# GetLogicalProcessInformation
# https://msdn.microsoft.com/en-us/library/ms683194(v=vs.85).aspx

# SYSTEM_LOGICAL_PROCESSOR_INFORMATION
# https://msdn.microsoft.com/en-us/library/ms686694(v=vs.85).aspx

# LOGICAL_PROCESSOR_RELATIONSHIP enum (0 - 4, 0xffff)
# https://msdn.microsoft.com/2ada52f0-70ec-4146-9ef7-9af3b08996f9

# CACHE_DESCRIPTOR struct
# https://msdn.microsoft.com/38cfa605-831c-45ef-a99f-55f42b2b56e9

# PROCESSOR_CACHE_TYPE
# https://msdn.microsoft.com/23044f67-e944-43c2-8c75-3d2fba87cb3c

# C example
# https://msdn.microsoft.com/en-us/904d2d35-f419-4e8f-a689-f39ed926644c



'''
Module of Android API for plyer.notification.

.. versionadded:: 1.0.0

.. versionchanged:: 1.4.0
    Fixed notifications not displaying due to missing NotificationChannel
    required by Android Oreo 8.0+ (API 26+).

.. versionchanged:: 1.4.0
    Added simple toaster notification.

.. versionchanged:: 1.4.0
    Fixed notifications not displaying big icons properly.
    Added option for custom big icon via `icon`.
'''

from __future__ import unicode_literals
from android import python_act
from android.runnable import run_on_ui_thread
from jnius import autoclass, cast

from plyer.facades import Notification
from plyer.platforms.android import activity, SDK_INT

AndroidString = autoclass('java.lang.String')
Context = autoclass('android.content.Context')
NotificationBuilder = autoclass('android.app.Notification$Builder')
NotificationManager = autoclass('android.app.NotificationManager')
Drawable = autoclass("{}.R$drawable".format(activity.getPackageName()))
PendingIntent = autoclass('android.app.PendingIntent')
Intent = autoclass('android.content.Intent')
Toast = autoclass('android.widget.Toast')
BitmapFactory = autoclass('android.graphics.BitmapFactory')


class AndroidNotification(Notification):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of Android notification API.

    .. versionadded:: 1.0.0
    '''

    def __init__(self):
        self._ns = None
        self._channel_id = activity.getPackageName()

    def _get_notification_service(self):
        if not self._ns:
            self._ns = cast(NotificationManager, activity.getSystemService(
                Context.NOTIFICATION_SERVICE
            ))
        return self._ns

    def _build_notification_channel(self, name):
        '''
        Create a NotificationChannel using channel id of the application
        package name (com.xyz, org.xyz, ...) and channel name same as the
        provided notification title if the API is high enough, otherwise
        do nothing.

        .. versionadded:: 1.4.0
        '''

        if SDK_INT < 26:
            return

        channel = autoclass('android.app.NotificationChannel')

        app_channel = channel(
            self._channel_id, name, NotificationManager.IMPORTANCE_DEFAULT
        )
        self._get_notification_service().createNotificationChannel(
            app_channel
        )
        return app_channel

    @run_on_ui_thread
    def _toast(self, message):
        '''
        Display a popup-like small notification at the bottom of the screen.

        .. versionadded:: 1.4.0
        '''
        Toast.makeText(
            activity,
            cast('java.lang.CharSequence', AndroidString(message)),
            Toast.LENGTH_LONG
        ).show()

    @staticmethod
    def _set_icons(notification, icon=None):
        '''
        Set the small application icon displayed at the top panel together with
        WiFi, battery percentage and time and the big optional icon (preferably
        PNG format with transparent parts) displayed directly in the
        notification body.

        .. versionadded:: 1.4.0
        '''
        app_icon = Drawable.icon
        notification.setSmallIcon(app_icon)

        bitmap_icon = app_icon
        if icon is not None:
            bitmap_icon = BitmapFactory.decodeFile(icon)
            notification.setLargeIcon(bitmap_icon)
        elif icon == '':
            # we don't want the big icon set,
            # only the small one in the top panel
            pass
        else:
            bitmap_icon = BitmapFactory.decodeResource(
                python_act.getResources(), app_icon
            )
            notification.setLargeIcon(bitmap_icon)

    def _build_notification(self, title):
        '''
        .. versionadded:: 1.4.0
        '''
        if SDK_INT < 26:
            noti = NotificationBuilder(activity)
        else:
            self._channel = self._build_notification_channel(title)
            noti = NotificationBuilder(activity, self._channel_id)
        return noti

    @staticmethod
    def _set_open_behavior(notification):
        '''
        Open the source application when user opens the notification.

        .. versionadded:: 1.4.0
        '''

        # create Intent that navigates back to the application
        app_context = activity.getApplication().getApplicationContext()
        notification_intent = Intent(app_context, python_act)

        # set flags to run our application Activity
        notification_intent.setFlags(Intent.FLAG_ACTIVITY_SINGLE_TOP)
        notification_intent.setAction(Intent.ACTION_MAIN)
        notification_intent.addCategory(Intent.CATEGORY_LAUNCHER)

        # get our application Activity
        pending_intent = PendingIntent.getActivity(
            app_context, 0, notification_intent, 0
        )

        notification.setContentIntent(pending_intent)
        notification.setAutoCancel(True)

    def _open_notification(self, notification):
        if SDK_INT >= 16:
            notification = notification.build()
        else:
            notification = notification.getNotification()

        self._get_notification_service().notify(0, notification)

    def _notify(self, **kwargs):
        noti = None
        message = kwargs.get('message').encode('utf-8')
        ticker = kwargs.get('ticker').encode('utf-8')
        title = AndroidString(
            kwargs.get('title', '').encode('utf-8')
        )
        icon = kwargs.get('app_icon')

        # decide whether toast only or proper notification
        if kwargs.get('toast'):
            self._toast(message)
            return
        else:
            noti = self._build_notification(title)

        # set basic properties for notification
        noti.setContentTitle(title)
        noti.setContentText(AndroidString(message))
        noti.setTicker(AndroidString(ticker))

        # set additional flags for notification
        self._set_icons(noti, icon=icon)
        self._set_open_behavior(noti)

        # launch
        self._open_notification(noti)


def instance():
    '''
    Instance for facade proxy.
    '''
    return AndroidNotification()



'''
Module of iOS API for plyer.email.
'''

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from plyer.facades import Email
from pyobjus import autoclass, objc_str  # pylint: disable=import-error
from pyobjus.dylib_manager import load_framework  # pylint:disable=import-error

load_framework('/System/Library/Frameworks/UIKit.framework')

NSURL = autoclass('NSURL')
NSString = autoclass('NSString')
UIApplication = autoclass('UIApplication')


class IOSEmail(Email):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of iOS battery API.
    '''

    def _send(self, **kwargs):
        recipient = kwargs.get('recipient')
        subject = kwargs.get('subject')
        text = kwargs.get('text')

        uri = "mailto:"
        if recipient:
            uri += str(recipient)
        if subject:
            uri += "?" if "?" not in uri else "&"
            uri += "subject="
            uri += quote(str(subject))
        if text:
            uri += "?" if "?" not in uri else "&"
            uri += "body="
            uri += quote(str(text))

        nsurl = NSURL.alloc().initWithString_(objc_str(uri))

        UIApplication.sharedApplication().openURL_(nsurl)


def instance():
    '''
    Instance for facade proxy.
    '''
    return IOSEmail()



# -- coding: utf-8 --
'''
Module of Windows API for creating taskbar balloon tip
notification in the taskbar's tray notification area.
'''

__all__ = ('WindowsBalloonTip', 'balloon_tip')


import time
import ctypes
from threading import RLock

from plyer.compat import PY2
from plyer.platforms.win.libs import win_api_defs


WS_OVERLAPPED = 0x00000000
WS_SYSMENU = 0x00080000
WM_DESTROY = 2
CW_USEDEFAULT = 8

LR_LOADFROMFILE = 16
LR_DEFAULTSIZE = 0x0040
IDI_APPLICATION = 32512
IMAGE_ICON = 1

NOTIFYICON_VERSION_4 = 4
NIM_ADD = 0
NIM_MODIFY = 1
NIM_DELETE = 2
NIM_SETVERSION = 4
NIF_MESSAGE = 1
NIF_ICON = 2
NIF_TIP = 4
NIF_INFO = 0x10
NIIF_USER = 4
NIIF_LARGE_ICON = 0x20


class WindowsBalloonTip(object):
    '''
    Implementation of balloon tip notifications through Windows API.

    * Register Window class name:
      https://msdn.microsoft.com/en-us/library/windows/desktop/ms632596.aspx
    * Create an overlapped window using the registered class.
      - It's hidden everywhere in GUI unless ShowWindow(handle, SW_SHOW)
        function is called.
    * Show/remove a tray icon and a balloon tip notification.

    Each instance is a separate notification with different parameters.
    Can be used with Threads.
    '''

    _class_atom = 0
    _wnd_class_ex = None
    _hwnd = None
    _hicon = None
    _balloon_icon = None
    _notify_data = None
    _count = 0
    _lock = RLock()

    @staticmethod
    def _get_unique_id():
        '''
        Keep track of each created balloon tip notification names,
        so that they can be easily identified even from outside.

        Make sure the count is shared between all the instances
        i.e. use a lock, so that _count class variable is incremented
        safely when using balloon tip notifications with Threads.
        '''

        WindowsBalloonTip._lock.acquire()
        val = WindowsBalloonTip._count
        WindowsBalloonTip._count += 1
        WindowsBalloonTip._lock.release()
        return val

    def __init__(self, title, message, app_name, app_icon='',
                 timeout=10, **kwargs):
        # pylint: disable=unused-argument,too-many-arguments
        '''
        The app_icon parameter, if given, is an .ICO file.
        '''

        wnd_class_ex = win_api_defs.get_WNDCLASSEXW()
        class_name = 'PlyerTaskbar' + str(WindowsBalloonTip._get_unique_id())

        if PY2:
            class_name = class_name.decode('utf8')
        wnd_class_ex.lpszClassName = class_name

        # keep ref to it as long as window is alive
        wnd_class_ex.lpfnWndProc = win_api_defs.WindowProc(
            win_api_defs.DefWindowProcW
        )
        wnd_class_ex.hInstance = win_api_defs.GetModuleHandleW(None)
        if wnd_class_ex.hInstance is None:
            raise Exception('Could not get windows module instance.')

        class_atom = win_api_defs.RegisterClassExW(wnd_class_ex)
        if class_atom == 0:
            raise Exception('Could not register the PlyerTaskbar class.')

        self._class_atom = class_atom
        self._wnd_class_ex = wnd_class_ex

        # create window
        self._hwnd = win_api_defs.CreateWindowExW(
            # dwExStyle, lpClassName, lpWindowName, dwStyle
            0, class_atom, '', WS_OVERLAPPED,
            # x, y, nWidth, nHeight
            0, 0, CW_USEDEFAULT, CW_USEDEFAULT,
            # hWndParent, hMenu, hInstance, lpParam
            None, None, wnd_class_ex.hInstance, None
        )
        if self._hwnd is None:
            raise Exception('Could not get create window.')
        win_api_defs.UpdateWindow(self._hwnd)

        # load .ICO file for as balloon tip and tray icon
        if app_icon:
            icon_flags = LR_LOADFROMFILE | LR_DEFAULTSIZE
            hicon = win_api_defs.LoadImageW(
                None, app_icon, IMAGE_ICON, 0, 0, icon_flags
            )

            if hicon is None:
                raise Exception('Could not load icon {}'.format(app_icon))
            self._balloon_icon = self._hicon = hicon
        else:
            self._hicon = win_api_defs.LoadIconW(
                None,
                ctypes.cast(IDI_APPLICATION, win_api_defs.LPCWSTR)
            )

        # show the notification
        self.notify(title, message, app_name)
        if timeout:
            time.sleep(timeout)

    def __del__(self):
        '''
        Clean visible parts of the notification object, then free all resources
        allocated for creating the nofitication Window and icon.
        '''
        self.remove_notify()
        if self._hicon is not None:
            win_api_defs.DestroyIcon(self._hicon)
        if self._wnd_class_ex is not None:
            win_api_defs.UnregisterClassW(
                self._class_atom,
                self._wnd_class_ex.hInstance
            )
        if self._hwnd is not None:
            win_api_defs.DestroyWindow(self._hwnd)

    def notify(self, title, message, app_name):
        '''
        Displays a balloon in the systray. Can be called multiple times
        with different parameter values.
        '''
        # remove previous visible balloon tip nofitication if available
        self.remove_notify()

        # add icon and messages to window
        hicon = self._hicon
        flags = NIF_TIP | NIF_INFO
        icon_flag = 0

        if hicon is not None:
            flags |= NIF_ICON

            # if icon is default app's one, don't display it in message
            if self._balloon_icon is not None:
                icon_flag = NIIF_USER | NIIF_LARGE_ICON

        notify_data = win_api_defs.get_NOTIFYICONDATAW(
            0, self._hwnd,
            id(self), flags, 0, hicon, app_name, 0, 0, message,
            NOTIFYICON_VERSION_4, title, icon_flag, win_api_defs.GUID(),
            self._balloon_icon
        )

        self._notify_data = notify_data
        if not win_api_defs.Shell_NotifyIconW(NIM_ADD, notify_data):
            raise Exception('Shell_NotifyIconW failed.')
        if not win_api_defs.Shell_NotifyIconW(NIM_SETVERSION,
                                              notify_data):
            raise Exception('Shell_NotifyIconW failed.')

    def remove_notify(self):
        '''
        Removes the notify balloon, if displayed.
        '''
        if self._notify_data is not None:
            win_api_defs.Shell_NotifyIconW(NIM_DELETE, self._notify_data)
            self._notify_data = None


def balloon_tip(**kwargs):
    '''
    Instance for balloon tip notification implementation.
    '''
    WindowsBalloonTip(**kwargs)



import ctypes
import os


def NSIterateSearchPaths(directory):
    LibraryPath = ("/System/Library/Frameworks/CoreFoundation.framework/"
                   "Versions/A/CoreFoundation")
    CoreFound = ctypes.cdll.LoadLibrary(LibraryPath)
    NSStartSearchPathEnumeration = CoreFound.NSStartSearchPathEnumeration
    NSGetNextSearchPathEnumeration = CoreFound.NSGetNextSearchPathEnumeration
    PATH_MAX = os.pathconf('/', os.pathconf_names['PC_PATH_MAX'])
    PATH_ENCODING = 'utf8'
    path_buffer = ctypes.create_string_buffer(PATH_MAX)
    # paths = []  <- fixme, possible list of paths in directory
    state = NSStartSearchPathEnumeration(directory, 1)
    while True:
        state = NSGetNextSearchPathEnumeration(state, path_buffer)
        if state == 0:
            break
        path = os.path.expanduser(path_buffer.value.decode(PATH_ENCODING))
        return path



'''
iOS Spatial Orientation
-----------------------

'''

from plyer.facades import SpatialOrientation
from pyobjus import autoclass


class iOSSpatialOrientation(SpatialOrientation):

    def __init__(self):
        self.bridge = autoclass('bridge').alloc().init()
        self.bridge.motionManager.setDeviceMotionUpdateInterval_(0.1)

    def _enable_listener(self):
        self.bridge.startDeviceMotion()

    def _disable_listener(self):
        self.bridge.stopDeviceMotion()

    def _get_orientation(self):
        return (
            self.bridge.sp_yaw,
            self.bridge.sp_pitch,
            self.bridge.sp_roll)


def instance():
    return iOSSpatialOrientation()



'''
Storage Path
============

The StorgePath API can be used to gain access to standard storage locations
across platforms such as home directory, root directory, external storage
directory, documents, downloads, etc.

The :class:`StoragePath` provides access to public methods to access standard
storage locations.

Simple Examples
---------------

To get the path of user's home directory::

    >>> from plyer import storagepath
    >>> storagepath.get_home_dir()

To get the path of standard downloads directory::

    >>> from plyer import storagepath
    >>> storagepath.get_downloads_dir()

To get the path of directory holding application files::

    >>> from plyer import storagepath
    >>> storagepath.get_application_dir()

'''


class StoragePath(object):
    '''
    StoragePath facade.
    '''

    def get_home_dir(self):
        '''
        Get the path of home directory of current user.
        '''
        return self._get_home_dir()

    def get_external_storage_dir(self):
        '''
        Get the path of primary shared or external storage directory.
        '''
        return self._get_external_storage_dir()

    def get_sdcard_dir(self):
        '''
        Get the path of external SD card.

        .. versionadded:: 1.4.0
        '''
        return self._get_sdcard_dir()

    def get_root_dir(self):
        '''
        Get the path of root of the "system" partition holding the core OS.
        '''
        return self._get_root_dir()

    def get_documents_dir(self):
        '''
        Get the path of standard directory in which to place documents that
        have been created by the user.
        '''
        return self._get_documents_dir()

    def get_downloads_dir(self):
        '''
        Get the path of standard directory in which to place files that have
        been downloaded by the user.
        '''
        return self._get_downloads_dir()

    def get_videos_dir(self):
        '''
        Get the path of standard directory in which to place videos that are
        available to the user.
        '''
        return self._get_videos_dir()

    def get_music_dir(self):
        '''
        Get the path of standard directory in which to place any audio files
        that should be in the regular list of music for the user.
        '''
        return self._get_music_dir()

    def get_pictures_dir(self):
        '''
        Standard directory in which to place pictures that are available to
        the user.
        '''
        return self._get_pictures_dir()

    def get_application_dir(self):
        '''
        Get the path of the directory holding application files.
        '''
        return self._get_application_dir()

    # private

    def _get_home_dir(self):
        raise NotImplementedError()

    def _get_external_storage_dir(self):
        raise NotImplementedError()

    def _get_sdcard_dir(self):
        raise NotImplementedError()

    def _get_root_dir(self):
        raise NotImplementedError()

    def _get_documents_dir(self):
        raise NotImplementedError()

    def _get_downloads_dir(self):
        raise NotImplementedError()

    def _get_videos_dir(self):
        raise NotImplementedError()

    def _get_music_dir(self):
        raise NotImplementedError()

    def _get_pictures_dir(self):
        raise NotImplementedError()

    def _get_application_dir(self):
        raise NotImplementedError()



'''
iOS Gyroscope
---------------------
'''

from plyer.facades import Gyroscope
from pyobjus import autoclass

from pyobjus.dylib_manager import load_framework

load_framework('/System/Library/Frameworks/UIKit.framework')
UIDevice = autoclass('UIDevice')

device = UIDevice.currentDevice()


class IosGyroscope(Gyroscope):

    def __init__(self):
        super(IosGyroscope, self).__init__()
        self.bridge = autoclass('bridge').alloc().init()

        if int(device.systemVersion.UTF8String().split('.')[0]) <= 4:
            self.bridge.motionManager.setGyroscopeUpdateInterval_(0.1)
        else:
            self.bridge.motionManager.setGyroUpdateInterval_(0.1)

        self.bridge.motionManager.setDeviceMotionUpdateInterval_(0.1)

    def _enable(self):
        self.bridge.startGyroscope()
        self.bridge.startDeviceMotion()

    def _disable(self):
        self.bridge.stopGyroscope()
        self.bridge.stopDeviceMotion()

    def _get_orientation(self):
        return (
            self.bridge.rotation_rate_x,
            self.bridge.rotation_rate_y,
            self.bridge.rotation_rate_z)

    def _get_rotation_uncalib(self):
        return (
            self.bridge.gy_x,
            self.bridge.gy_y,
            self.bridge.gy_z,
            self.bridge.gy_x - self.bridge.rotation_rate_x,
            self.bridge.gy_y - self.bridge.rotation_rate_y,
            self.bridge.gy_z - self.bridge.rotation_rate_z)


def instance():
    return IosGyroscope()



'''
Module of Android API for plyer.battery.
'''

from jnius import autoclass, cast  # pylint: disable=no-name-in-module
from plyer.platforms.android import activity
from plyer.facades import Battery

Intent = autoclass('android.content.Intent')
BatteryManager = autoclass('android.os.BatteryManager')
IntentFilter = autoclass('android.content.IntentFilter')


class AndroidBattery(Battery):
    '''
    Implementation of Android battery API.
    '''

    def _get_state(self):
        status = {"isCharging": None, "percentage": None}

        ifilter = IntentFilter(Intent.ACTION_BATTERY_CHANGED)

        battery_status = cast(
            'android.content.Intent',
            activity.registerReceiver(None, ifilter)
        )

        query = battery_status.getIntExtra(BatteryManager.EXTRA_STATUS, -1)
        is_charging = query == BatteryManager.BATTERY_STATUS_CHARGING
        is_full = query == BatteryManager.BATTERY_STATUS_FULL

        level = battery_status.getIntExtra(BatteryManager.EXTRA_LEVEL, -1)
        scale = battery_status.getIntExtra(BatteryManager.EXTRA_SCALE, -1)
        percentage = (level / float(scale)) * 100

        status['isCharging'] = is_charging or is_full
        status['percentage'] = percentage

        return status


def instance():
    '''
    Instance for facade proxy.
    '''
    return AndroidBattery()



'''
Battery
=======

The :class:`Battery` provides information about the battery of your device.

.. note::
        On Android the `BATTERY_STATS` permission is needed.

Simple Example
---------------

To get battery status::

    >>> from plyer import battery
    >>> battery.status
    {'percentage': 82.0, 'isCharging': False}

Supported Platforms
-------------------
Android, iOS, Windows, OS X, Linux

'''


class Battery(object):
    '''
    Battery info facade.
    '''

    @property
    def status(self):
        '''
        Property that contains a dict with the following fields:
             * **isCharging** *(bool)*: Battery is charging
             * **percentage** *(float)*: Battery charge remaining

            .. warning::
                If any of the fields is not readable, it is set as
                None.
        '''
        return self.get_state()

    def get_state(self):
        '''
        Public method for filling battery.status via platform-specific
        API in plyer.platforms.
        '''
        return self._get_state()

    # private

    def _get_state(self):
        raise NotImplementedError()



'''
IrBlaster
============

The :class:`IrBlaster` provides access to public methods by which your device
can act as a remote and could be used to control your TV, AC, Music Player,
Projectors, Set top box or anything that can be controlled by a remote.

.. note::
    - On Android your app needs the TRANSMIT_IR permission which allows an
    application to use the device's IR transmitter, If available.

Simple Examples
---------------

To get transmit an IR sequence::

    >>> from plyer import irblaster
    >>> irblaster.transmit(frequency, pattern, mode)

To get frequencies::

    >>> irblaster.frequencies

To check if IrBlaster exists::

    >>> irblaster.exists()
    True/False

Supported Platforms
-------------------
Android

'''


class IrBlaster(object):
    '''
    Infrared blaster facade.
    '''

    @staticmethod
    def periods_to_microseconds(frequency, pattern):
        '''
        Convert a pattern from period counts to microseconds.
        '''
        period = 1000000. / frequency
        return [period * x for x in pattern]

    @staticmethod
    def microseconds_to_periods(frequency, pattern):
        '''
        Convert a pattern from microseconds to period counts.
        '''
        period = 1000000. / frequency
        return [x / period for x in pattern]

    @property
    def frequencies(self):
        '''
        Property which contains a list of frequency ranges
           supported by the device in the form:

           [(from1, to1),
            (from2, to2),
            ...
            (fromN, toN)]
        '''
        return self.get_frequencies()

    def get_frequencies(self):
        return self._get_frequencies()

    def transmit(self, frequency, pattern, mode='period'):
        '''
        Transmit an IR sequence.

        :parameters:
            `frequency`: int
                Carrier frequency for the IR transmission.
            `pattern`: list[int]
                Burst pair pattern to transmit.
            `mode`: str, defaults to 'period'
                Specifies the format of the pattern values.
                Can be 'period' or 'microseconds'.
        '''
        return self._transmit(frequency, pattern, mode)

    def exists(self):
        '''
        Check if the device has an infrared emitter.
        '''
        return self._exists()

    # private

    def _get_frequencies(self):
        raise NotImplementedError()

    def _transmit(self, frequency, pattern, mode):
        raise NotImplementedError()

    def _exists(self):
        raise NotImplementedError()



'''
Android Brightness
------------------
'''

from jnius import autoclass
from plyer.facades import Brightness
from android import mActivity

System = autoclass('android.provider.Settings$System')


class AndroidBrightness(Brightness):

    def _current_level(self):

        System.putInt(
            mActivity.getContentResolver(),
            System.SCREEN_BRIGHTNESS_MODE,
            System.SCREEN_BRIGHTNESS_MODE_MANUAL)
        cr_level = System.getInt(
            mActivity.getContentResolver(),
            System.SCREEN_BRIGHTNESS)
        return (cr_level / 255.) * 100

    def _set_level(self, level):
        System.putInt(
            mActivity.getContentResolver(),
            System.SCREEN_BRIGHTNESS,
            (level / 100.) * 255)


def instance():
    return AndroidBrightness()



'''
Module of Android API for plyer.bluetooth.
'''

from jnius import autoclass
from plyer.platforms.android import activity
from plyer.facades import Bluetooth

Global = autoclass('android.provider.Settings$Global')


class AndroidBluetooth(Bluetooth):
    '''
    Implementation of Android Bluetooth API.
    '''

    def _get_info(self):
        bluetooth_enabled = Global.getString(
            activity.getContentResolver(),
            Global.BLUETOOTH_ON
        )
        status = 'off'
        if bluetooth_enabled:
            status = 'on'
        return status


def instance():
    '''
    Instance for facade proxy.
    '''
    return AndroidBluetooth()



'''
CPU
===

Simple Example
---------------

To get CPU count::
    >>> from plyer import cpu
    >>> # 1 socket, 1 core per socket, 2 threads per core
    >>> cpu.sockets   # 1 CPU socket (or slot)
    1
    >>> cpu.physical  # 1 CPU socket * 1 core per socket
    1
    >>> cpu.logical   # 1 CPU socket * 1 core per socket * 2 threads per core
    2

Supported Platforms
-------------------

MacOS
Linux
Windows
'''


class CPU(object):
    '''
    Facade providing info about sockets, physical and logical
    number of processors.
    '''

    @property
    def sockets(self):
        '''
        Property that contains the count of CPU sockets.
        '''
        return self._sockets()

    @property
    def physical(self):
        '''
        Property that contains the total number of physical cores
        (max core count) in the system.

        .. note:: `sockets * cores per socket`
        '''
        return self._physical()

    @property
    def logical(self):
        '''
        Property that contains the total number of logical cores
        (max thread count) in the system.

        .. note:: `sockets * cores per socket * threads per core`
        '''
        return self._logical()

    @property
    def cache(self):
        '''
        Property that contains the count of L1, L2, L3 caches in the system
        as a dictionary `{'L1': int, 'L2': int, 'L3': int}`.
        '''
        return self._cache()

    @property
    def numa(self):
        '''
        Property that contains the count of NUMA nodes in the system.

        .. note:: https://en.wikipedia.org/wiki/Non-uniform_memory_access
        '''
        return self._numa()

    # private

    def _sockets(self):
        raise NotImplementedError()

    def _physical(self):
        raise NotImplementedError()

    def _logical(self):
        raise NotImplementedError()

    def _cache(self):
        raise NotImplementedError()

    def _numa(self):
        raise NotImplementedError()



import subprocess as sb
from plyer.facades import Orientation


class LinuxOrientation(Orientation):

    def _set_landscape(self, **kwargs):
        self.rotate = 'normal'
        self.screen = sb.check_output(
            "xrandr -q | grep ' connected' |  head -n 1 | cut -d ' ' -f1",
            shell=True
        )
        self.screen = self.screen.decode('utf-8').split('\n')[0]
        sb.call(["xrandr", "--output", self.screen, "--rotate", self.rotate])

    def _set_portrait(self, **kwargs):
        self.rotate = 'left'
        self.screen = sb.check_output(
            "xrandr -q | grep ' connected' |  head -n 1 | cut -d ' ' -f1",
            shell=True
        )
        self.screen = self.screen.decode('utf-8').split('\n')[0]
        sb.call(["xrandr", "--output", self.screen, "--rotate", self.rotate])


def instance():
    return LinuxOrientation()



'''
Module of Linux API for plyer.battery.
'''

from math import floor
from os import environ
from os.path import exists, join
from subprocess import Popen, PIPE
from plyer.facades import Battery
from plyer.utils import whereis_exe


class LinuxBattery(Battery):
    '''
    Implementation of Linux battery API via accessing the sysclass power_supply
    path from the kernel.
    '''

    def _get_state(self):
        status = {"isCharging": None, "percentage": None}

        kernel_bat_path = join('/sys', 'class', 'power_supply', 'BAT0')
        uevent = join(kernel_bat_path, 'uevent')

        with open(uevent) as fle:
            lines = [
                line.decode('utf-8').strip()
                for line in fle.readlines()
            ]
        output = {
            line.split('=')[0]: line.split('=')[1]
            for line in lines
        }

        is_charging = output['POWER_SUPPLY_STATUS'] == 'Charging'
        total = float(output['POWER_SUPPLY_CHARGE_FULL'])
        now = float(output['POWER_SUPPLY_CHARGE_NOW'])

        capacity = floor(now / total * 100)

        status['percentage'] = capacity
        status['isCharging'] = is_charging
        return status


class UPowerBattery(Battery):
    '''
    Implementation of UPower battery API.
    '''

    def _get_state(self):
        # if no LANG specified, return empty string
        old_lang = environ.get('LANG', '')
        environ['LANG'] = 'C'
        status = {"isCharging": None, "percentage": None}

        # We are supporting only one battery now
        # this will fail if there is no object with such path,
        # however it's safer than 'upower -d' which provides
        # multiple unrelated 'state' and 'percentage' keywords
        dev = "/org/freedesktop/UPower/devices/battery_BAT0"
        upower_process = Popen(
            ["upower", "--show-info", dev],
            stdout=PIPE
        )
        output = upower_process.communicate()[0].decode()
        environ['LANG'] = old_lang
        if not output:
            return status
        state = percentage = None

        for line in output.splitlines():
            if 'state' in line:
                state = line.rpartition(':')[-1].strip()

            if 'percentage' in line:
                percentage = line.rpartition(':')[-1].strip()[:-1]

                # switching decimal comma to dot
                # (different LC_NUMERIC locale)
                percentage = float(
                    percentage.replace(',', '.')
                )

        if state:
            status['isCharging'] = state == "charging"
        status['percentage'] = percentage
        return status


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('upower'):
        return UPowerBattery()
    sys.stderr.write("upower not found.")

    if exists(join('/sys', 'class', 'power_supply', 'BAT0')):
        return LinuxBattery()
    return Battery()



'''
Android file chooser
--------------------

Android runs ``Activity`` asynchronously via pausing our ``PythonActivity``
and starting a new one in the foreground. This means
``AndroidFileChooser._open_file()`` will always return the default value of
``AndroidFileChooser.selection`` i.e. ``None``.

After the ``Activity`` (for us it's the file chooser ``Intent``) is completed,
Android moves it to the background (or destroys or whatever is implemented)
and pushes ``PythonActivity`` to the foreground.

We have a custom listener for ``android.app.Activity.onActivityResult()``
via `android` package from `python-for-android` recipe,
``AndroidFileChooser._on_activity_result()`` which is called independently of
any our action (we may call anything from our application in Python and this
handler will be called nevertheless on each ``android.app.Activity`` result
in the system).

In the handler we check if the ``request_code`` matches the code passed to the
``Context.startActivityForResult()`` i.e. if the result from
``android.app.Activity`` is indeed meant for our ``PythonActivity`` and then we
proceed.

Since the ``android.app.Activity.onActivityResult()`` is the only way for us
to intercept the result and we have a handler bound via ``android`` package,
we need to get the path/file/... selection to the user the same way.

Threading + ``Thread.join()`` or ``time.sleep()`` or any other kind of waiting
for the result is not an option because:

1) ``android.app.Activity.onActivityResult()`` might remain unexecuted if
the launched file chooser activity does not return the result (``Activity``
dies/freezes/etc).

2) Thread will be still waiting for the result e.g. an update of a value or
to actually finish, however the result from the call of
``AndroidFileChooser._open_file()`` will be returned nevertheless and anything
using that result will use an incorrect one i.e. the default value of
``AndroidFilechooser.selection`` (``None``).

.. versionadded:: 1.4.0
'''

from __future__ import unicode_literals
from os.path import join, basename
from random import randint

from android import activity, mActivity
from jnius import autoclass, cast, JavaException
from plyer.facades import FileChooser
from plyer import storagepath


String = autoclass('java.lang.String')
Intent = autoclass('android.content.Intent')
Activity = autoclass('android.app.Activity')
DocumentsContract = autoclass('android.provider.DocumentsContract')
ContentUris = autoclass('android.content.ContentUris')
Uri = autoclass('android.net.Uri')
Long = autoclass('java.lang.Long')
IMedia = autoclass('android.provider.MediaStore$Images$Media')
VMedia = autoclass('android.provider.MediaStore$Video$Media')
AMedia = autoclass('android.provider.MediaStore$Audio$Media')


class AndroidFileChooser(FileChooser):
    '''
    FileChooser implementation for Android using
    the built-in file browser via Intent.

    .. versionadded:: 1.4.0
    '''

    # filechooser activity <-> result pair identification
    select_code = None

    # default selection value
    selection = None

    def __init__(self, *args, **kwargs):
        super(AndroidFileChooser, self).__init__(*args, **kwargs)
        self.select_code = randint(123456, 654321)
        self.selection = None

        # bind a function for a response from filechooser activity
        activity.bind(on_activity_result=self._on_activity_result)

    @staticmethod
    def _handle_selection(selection):  # pylint: disable=method-hidden
        '''
        Dummy placeholder for returning selection from
        ``android.app.Activity.onActivityResult()``.

        .. versionadded:: 1.4.0
        '''
        return selection

    def _open_file(self, **kwargs):
        '''
        Running Android Activity is non-blocking and the only call
        that blocks is onActivityResult running in GUI thread

        .. versionadded:: 1.4.0
        '''

        # set up selection handler
        # startActivityForResult is async
        # onActivityResult is sync
        self._handle_selection = kwargs.pop(
            'on_selection', self._handle_selection
        )

        # create Intent for opening
        file_intent = Intent(Intent.ACTION_GET_CONTENT)
        file_intent.setType('*/*')
        file_intent.addCategory(
            Intent.CATEGORY_OPENABLE
        )

        # start a new activity from PythonActivity
        # which creates a filechooser via intent
        mActivity.startActivityForResult(
            Intent.createChooser(file_intent, cast(
                'java.lang.CharSequence',
                String("FileChooser")
            )),
            self.select_code
        )

    def _on_activity_result(self, request_code, result_code, data):
        '''
        Listener for ``android.app.Activity.onActivityResult()`` assigned
        via ``android.activity.bind()``.

        .. versionadded:: 1.4.0
        '''

        # not our response
        if request_code != self.select_code:
            return

        if result_code != Activity.RESULT_OK:
            # The action had been cancelled.
            return

        selection = self._resolve_uri(data.getData()) or []

        # return value to object
        self.selection = [selection]
        # return value via callback
        self._handle_selection([selection])

    @staticmethod
    def _handle_external_documents(uri):
        '''
        Selection from the system filechooser when using ``Phone``
        or ``Internal storage`` or ``SD card`` option from menu.

        .. versionadded:: 1.4.0
        '''

        file_id = DocumentsContract.getDocumentId(uri)
        file_type, file_name = file_id.split(':')

        # internal SD card mostly mounted as a files storage in phone
        internal = storagepath.get_external_storage_dir()

        # external (removable) SD card i.e. microSD
        external = storagepath.get_sdcard_dir()
        external_base = basename(external)

        # resolve sdcard path
        sdcard = internal

        # because external might have /storage/.../1 or other suffix
        # and file_type might be only a part of the real folder in /storage
        if file_type in external_base or external_base in file_type:
            sdcard = external

        path = join(sdcard, file_name)
        return path

    @staticmethod
    def _handle_media_documents(uri):
        '''
        Selection from the system filechooser when using ``Images``
        or ``Videos`` or ``Audio`` option from menu.

        .. versionadded:: 1.4.0
        '''

        file_id = DocumentsContract.getDocumentId(uri)
        file_type, file_name = file_id.split(':')
        selection = '_id=?'

        if file_type == 'image':
            uri = IMedia.EXTERNAL_CONTENT_URI
        elif file_type == 'video':
            uri = VMedia.EXTERNAL_CONTENT_URI
        elif file_type == 'audio':
            uri = AMedia.EXTERNAL_CONTENT_URI
        return file_name, selection, uri

    @staticmethod
    def _handle_downloads_documents(uri):
        '''
        Selection from the system filechooser when using ``Downloads``
        option from menu. Might not work all the time due to:

        1) invalid URI:

        jnius.jnius.JavaException:
            JVM exception occurred: Unknown URI:
            content://downloads/public_downloads/1034

        2) missing URI / android permissions

        jnius.jnius.JavaException:
            JVM exception occurred:
            Permission Denial: reading
            com.android.providers.downloads.DownloadProvider uri
            content://downloads/all_downloads/1034 from pid=2532, uid=10455
            requires android.permission.ACCESS_ALL_DOWNLOADS,
            or grantUriPermission()

        Workaround:
            Selecting path from ``Phone`` -> ``Download`` -> ``<file>``
            (or ``Internal storage``) manually.

        .. versionadded:: 1.4.0
        '''

        # known locations, differ between machines
        downloads = [
            'content://downloads/public_downloads',
            'content://downloads/my_downloads',

            # all_downloads requires separate permission
            # android.permission.ACCESS_ALL_DOWNLOADS
            'content://downloads/all_downloads'
        ]

        file_id = DocumentsContract.getDocumentId(uri)
        try_uris = [
            ContentUris.withAppendedId(
                Uri.parse(down), Long.valueOf(file_id)
            )
            for down in downloads
        ]

        # try all known Download folder uris
        # and handle JavaExceptions due to different locations
        # for content:// downloads or missing permission
        path = None
        for down in try_uris:
            try:
                path = AndroidFileChooser._parse_content(
                    uri=down, projection=['_data'],
                    selection=None,
                    selection_args=None,
                    sort_order=None
                )

            except JavaException:
                import traceback
                traceback.print_exc()

            # we got a path, ignore the rest
            if path:
                break

        # alternative approach to Downloads by joining
        # all data items from Activity result
        if not path:
            for down in try_uris:
                try:
                    path = AndroidFileChooser._parse_content(
                        uri=down, projection=None,
                        selection=None,
                        selection_args=None,
                        sort_order=None,
                        index_all=True
                    )

                except JavaException:
                    import traceback
                    traceback.print_exc()

                # we got a path, ignore the rest
                if path:
                    break
        return path

    def _resolve_uri(self, uri):
        '''
        Resolve URI input from ``android.app.Activity.onActivityResult()``.

        .. versionadded:: 1.4.0
        '''

        uri_authority = uri.getAuthority()
        uri_scheme = uri.getScheme().lower()

        path = None
        file_name = None
        selection = None
        downloads = None

        # not a document URI, nothing to convert from
        if not DocumentsContract.isDocumentUri(mActivity, uri):
            return path

        if uri_authority == 'com.android.externalstorage.documents':
            return self._handle_external_documents(uri)

        # in case a user selects a file from 'Downloads' section
        # note: this won't be triggered if a user selects a path directly
        #       e.g.: Phone -> Download -> <some file>
        elif uri_authority == 'com.android.providers.downloads.documents':
            path = downloads = self._handle_downloads_documents(uri)

        elif uri_authority == 'com.android.providers.media.documents':
            file_name, selection, uri = self._handle_media_documents(uri)

        # parse content:// scheme to path
        if uri_scheme == 'content' and not downloads:
            path = self._parse_content(
                uri=uri, projection=['_data'], selection=selection,
                selection_args=[file_name], sort_order=None
            )

        # nothing to parse, file:// will return a proper path
        elif uri_scheme == 'file':
            path = uri.getPath()

        return path

    @staticmethod
    def _parse_content(  # pylint: disable=too-many-arguments
            uri, projection, selection, selection_args, sort_order,
            index_all=False
    ):
        '''
        Parser for ``content://`` URI returned by some Android resources.

        .. versionadded:: 1.4.0
        '''

        result = None
        resolver = mActivity.getContentResolver()
        read = Intent.FLAG_GRANT_READ_URI_PERMISSION
        write = Intent.FLAG_GRANT_READ_URI_PERMISSION
        persist = Intent.FLAG_GRANT_READ_URI_PERMISSION

        # grant permission for our activity
        mActivity.grantUriPermission(
            mActivity.getPackageName(),
            uri,
            read | write | persist
        )

        if not index_all:
            cursor = resolver.query(
                uri, projection, selection,
                selection_args, sort_order
            )

            idx = cursor.getColumnIndex(projection[0])
            if idx != -1 and cursor.moveToFirst():
                result = cursor.getString(idx)
        else:
            result = []
            cursor = resolver.query(
                uri, projection, selection,
                selection_args, sort_order
            )
            while cursor.moveToNext():
                for idx in range(cursor.getColumnCount()):
                    result.append(cursor.getString(idx))
            result = '/'.join(result)
        return result

    def _file_selection_dialog(self, **kwargs):
        mode = kwargs.pop('mode', None)
        if mode == 'open':
            self._open_file(**kwargs)


def instance():
    return AndroidFileChooser()



'''
Linux Brightness
----------------

'''

from plyer.facades import Brightness
import subprocess
import os


class LinuxBrightness(Brightness):

    def __init__(self):
        if os.system("which xbacklight"):
            msg = ("It looks like 'xbacklight' is not installed. Try "
                   "installing it with your distribution's package manager.")
            raise Exception(msg)

    def _current_level(self):
        cr_level = subprocess.check_output(["xbacklight", "-get"])
        return str(cr_level)

    def _set_level(self, level):
        subprocess.call(["xbacklight", "-set", str(level)])


def instance():
    return LinuxBrightness()



'''
Gyroscope
============

The gyroscope measures the rate of rotation around a device's x, y,
and z axis.

The :class:`Gyroscope` provides access to public methods to
use gyroscope of your device.

Simple Examples
---------------

To enable gyroscope::

    >>> from plyer import gyroscope
    >>> gyroscope.enable()

To disable gyroscope::

    >>> gyroscope.disable()

To get the rate of rotation along the three axes::

    >>> gyroscope.rotation
    (-0.0034587313421070576, -0.0073830625042319298, 0.0046892408281564713)

To get the uncalibrated rate of rotation along the three axes along with the
drift compensation::

    >>> gyroscope.rotation_uncalib
    ()
    where the first three values show the rate of rotation w/o drift
    compensation and the last three show the estimated drift along the three
    axes.

Supported Platforms
-------------------
Android, iOS

'''


class Gyroscope(object):
    '''
    Gyroscope facade.

    .. versionadded:: 1.3.1
    '''

    @property
    def rotation(self):
        '''
        Property that returns the rate of rotation around the device's local
        X, Y and Z axis.

        Along x-axis: angular speed around the X axis
        Along y-axis: angular speed around the Y axis
        Along z-axis: angular speed around the Z axis

        Returns (None, None, None) if no data is currently available.
        '''
        return self.get_orientation()

    @property
    def rotation_uncalib(self):
        '''
        Property that returns the current rate of rotation around the X, Y and
        Z axis. An estimation of the drift on each axis is reported as well.

        Along x-axis: angular speed (w/o drift compensation) around the X axis
        Along y-axis: angular speed (w/o drift compensation) around the Y axis
        Along z-axis: angular speed (w/o drift compensation) around the Z axis

        Along x-axis: estimated drift around X axis
        Along y-axis: estimated drift around Y axis
        Along z-axis: estimated drift around Z axis

        Returns (None, None, None, None, None, None) if no data is currently
        available.
        '''
        return self.get_rotation_uncalib()

    @property
    def orientation(self):
        '''
        WARNING:: This property is deprecated after API Level 8.
        Use `gyroscope.rotation` instead.

        Property that returns values of the current Gyroscope sensors, as
        a (x, y, z) tuple. Returns (None, None, None) if no data is currently
        available.
        '''
        return self.get_orientation()

    def enable(self):
        '''
        Activate the Gyroscope sensor.
        '''
        self._enable()

    def disable(self):
        '''
        Disable the Gyroscope sensor.
        '''
        self._disable()

    def get_orientation(self):
        return self._get_orientation()

    def get_rotation_uncalib(self):
        return self._get_rotation_uncalib()

    # private

    def _enable(self):
        raise NotImplementedError()

    def _disable(self):
        raise NotImplementedError()

    def _get_orientation(self):
        raise NotImplementedError()

    def _get_rotation_uncalib(self):
        raise NotImplementedError()



from time import sleep
from jnius import autoclass
from plyer.facades import TTS
from plyer.platforms.android import activity

Locale = autoclass('java.util.Locale')
TextToSpeech = autoclass('android.speech.tts.TextToSpeech')


class AndroidTextToSpeech(TTS):
    def _speak(self, **kwargs):
        tts = TextToSpeech(activity, None)

        tts.setLanguage(Locale.US)

        retries = 0  # First try rarely succeeds due to some timing issue
        message = kwargs.get('message')

        # first try for while loop
        speak_status = tts.speak(
            message, TextToSpeech.QUEUE_FLUSH, None
        )

        # -1 indicates error. Let's wait and then try again
        while retries < 100 and speak_status == -1:
            sleep(0.1)
            retries += 1
            speak_status = tts.speak(
                message, TextToSpeech.QUEUE_FLUSH, None
            )


def instance():
    return AndroidTextToSpeech()



'''
Module of Linux API for plyer.uniqueid.
'''

from os import environ
from subprocess import Popen, PIPE
from plyer.facades import UniqueID
from plyer.utils import whereis_exe


class LinuxUniqueID(UniqueID):
    '''
    Implementation of Linux uniqueid API.
    '''

    def _get_uid(self):
        old_lang = environ.get('LANG')
        environ['LANG'] = 'C'
        stdout = Popen(
            ["lshw", "-quiet"],
            stdout=PIPE, stderr=PIPE
        ).communicate()[0].decode('utf-8')

        output = u''
        for line in stdout.splitlines():
            if 'serial:' not in line:
                continue
            output = line
            break

        environ['LANG'] = old_lang or u''
        result = None

        if output:
            result = output.split()[1]
        return result


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('lshw'):
        return LinuxUniqueID()
    sys.stderr.write("lshw not found.")
    return UniqueID()



from plyer.facades import Keystore
from pyobjus import autoclass, objc_str

NSUserDefaults = autoclass('NSUserDefaults')


class IosKeystore(Keystore):

    def _set_key(self, servicename, key, value, **kwargs):
        NSUserDefaults.standardUserDefaults().setObject_forKey_(
            objc_str(value), objc_str(key))

    def _get_key(self, servicename, key, **kwargs):
        ret = NSUserDefaults.standardUserDefaults().stringForKey_(
            objc_str(key))
        if ret is not None:
            return ret.UTF8String()
        else:
            return ret


def instance():
    return IosKeystore()



# coding=utf-8


class SpatialOrientation(object):
    '''Spatial Orientation facade.

    Computes the device's orientation based on the rotation matrix.

    .. versionadded:: 1.3.1
    '''

    @property
    def orientation(self):
        '''Property that returns values of the current device orientation
        as a (azimuth, pitch, roll) tuple.

        Azimuth, angle of rotation about the -z axis. This value represents the
        angle between the device's y axis and the magnetic north pole.
        The range of values is -π to π.

        Pitch, angle of rotation about the x axis. This value represents the
        angle between a plane parallel to the device's screen and a plane
        parallel to the ground.
        The range of values is -π to π.

        Roll, angle of rotation about the y axis. This value represents the
        angle between a plane perpendicular to the device's screen and a plane
        perpendicular to the ground.
        The range of values is -π/2 to π/2.

        Returns (None, None, None) if no data is currently available.

        Supported Platforms:: Android
        '''
        return self._get_orientation() or (None, None, None)

    def _get_orientation(self):
        raise NotImplementedError()

    def enable_listener(self):
        '''Enable the orientation sensor.
        '''
        self._enable_listener()

    def _enable_listener(self, **kwargs):
        raise NotImplementedError()

    def disable_listener(self):
        '''Disable the orientation sensor.
        '''
        self._disable_listener()

    def _disable_listener(self, **kwargs):
        raise NotImplementedError()



'''
Wifi Facade.
=============

The :class:`Wifi` is to provide access to the wifi of your mobile/ desktop
devices.

It currently supports `connecting`, `disconnecting`, `scanning`, `getting
available wifi network list` and `getting network information`.

Simple examples
---------------

To enable/ turn on wifi scanning::

    >>> from plyer import wifi
    >>> wifi.start_scanning()

Once the wifi is enabled/ turned on, then this command starts to scan
all the nearby available wifi networks.

To get network info::

    >>> from plyer import wifi
    >>> wifi.start_scanning()
    >>> return wifi.get_network_info(name)

Returns network details of the network who's name/ssid is provided in the
`name` parameter.

To connect to a network::

    >>> from plyer import wifi
    >>> wifi.start_scanning()
    >>> wifi.connect(network, parameters)

This connects to the network who's name/ssid is provided under `network`
parameter and along with other necessary methods for connection
which depends upon platform to platform.

please visit following files for more details about requirements of
`paramaters` argument in `connect` method:

    plyer/platforms/win/wifi.py
    plyer/platforms/macosx/wifi.py
    plyer/platforms/win/wifi.py

To disconnect from wifi::

    >>> from plyer import wifi
    >>> wifi.disconnect()

This disconnects your device from any wifi network.

To get available wifi networks::

    >>> from plyer import wifi
    >>> wifi.start_scanning()
    >>> return wifi.get_available_wifi()

This returns all the available wifi networks near the device.

Supported Platforms
-------------------
Windows, OS X, Linux

Ex: 6
----------

from plyer import wifi
wifi.enable()

This enables wifi device.

Ex: 7
----------

from plyer import wifi
wifi.disable()

This disable wifi device
'''


class Wifi(object):
    '''
    Wifi Facade.
    '''

    def is_enabled(self):
        '''
        Return enabled status of WiFi hardware.
        '''
        return self._is_enabled()

    def is_connected(self, interface=None):
        '''
        Return connection state of WiFi interface.

        .. versionadded:: 1.4.0
        '''
        return self._is_connected(interface=interface)

    @property
    def interfaces(self):
        '''
        List all available WiFi interfaces.

        .. versionadded:: 1.4.0
        '''

        raise NotImplementedError()

    def start_scanning(self, interface=None):
        '''
        Turn on scanning.
        '''
        return self._start_scanning(interface=interface)

    def get_network_info(self, name):
        '''
        Return a dictionary of secified network.
        '''
        return self._get_network_info(name=name)

    def get_available_wifi(self):
        '''
        Returns a list of all the available wifi.
        '''
        return self._get_available_wifi()

    def connect(self, network, parameters, interface=None):
        '''
        Method to connect to some network.
        '''
        self._connect(
            network=network,
            parameters=parameters,
            interface=interface
        )

    def disconnect(self, interface=None):
        '''
        To disconnect from some network.
        '''
        self._disconnect(interface=interface)

    def enable(self):
        '''
        Wifi interface power state is set to "ON".
        '''
        self._enable()

    def disable(self):
        '''
        Wifi interface power state is set to "OFF".
        '''
        self._disable()

    # private

    def _is_enabled(self):
        raise NotImplementedError()

    def _is_connected(self, interface=None):
        raise NotImplementedError()

    def _start_scanning(self, interface=None):
        raise NotImplementedError()

    def _get_network_info(self, **kwargs):
        raise NotImplementedError()

    def _get_available_wifi(self):
        raise NotImplementedError()

    def _connect(self, **kwargs):
        raise NotImplementedError()

    def _disconnect(self, interface=None):
        raise NotImplementedError()

    def _enable(self):
        raise NotImplementedError()

    def _disable(self):
        raise NotImplementedError()



'''
IOS Call
----------
'''

from plyer.facades import Call
from pyobjus import autoclass, objc_str

NSURL = autoclass('NSURL')
NSString = autoclass('NSString')
UIApplication = autoclass('UIApplication')


class IOSCall(Call):

    def _makecall(self, **kwargs):
        tel = kwargs.get('tel')
        url = "tel://" + tel
        nsurl = NSURL.alloc().initWithString_(objc_str(url))

        UIApplication.sharedApplication().openURL_(nsurl)

    def _dialcall(self, **kwargs):
        pass
        # Not possible, Access not provided by iPhone SDK


def instance():
    return IOSCall()



# coding=utf-8
'''
Flash
=====

The :class:`Flash` provides access to public methods to use flash of your
device.

.. note::
    In android you need CAMERA, FLASHLIGHT permissions
    to access flash.

.. versionadded:: 1.2.5

This can be used to activate the flash of your camera on
Android and iOS.

Simple Examples
---------------

To turn on flash::

    >>> from plyer import flash
    >>> flash.on()

To turn off flash::

    >>> flash.off()

To release flash::

    >>> flash.release()

Supported Platforms
-------------------
Android, iOS

'''


class Flash(object):
    """
    Flash facade.
    """

    def on(self):
        """
        Activate the flash
        """
        self._on()

    def off(self):
        """
        Deactiavte the flash
        """
        self._off()

    def release(self):
        """
        Release any access to the Flash / Camera.
        Call this when you're done using the Flash.
        This will release the Camera, and stop any process.

        Next call to `_on` will reactivate it.
        """
        self._release()

    # private

    def _on(self):
        raise NotImplementedError()

    def _off(self):
        raise NotImplementedError()

    def _release(self):
        pass



'''
Audio
=====

The :class:`Audio` is used for recording audio.

Default path for recording is set in platform implementation.

.. note::
        On Android the `RECORD_AUDIO`, `WAKE_LOCK` permissions are needed.

Simple Examples
---------------

To get the file path::

    >>> audio.file_path
    '/sdcard/testrecorder.3gp'

To set the file path::

    >>> import os
    >>> current_list = os.listdir('.')
    ['/sdcard/testrecorder.3gp', '/sdcard/testrecorder1.3gp',
    '/sdcard/testrecorder2.3gp', '/sdcard/testrecorder3.3gp']
    >>> file_path = current_list[2]
    >>> audio.file_path = file_path

To start recording::

    >>> from plyer import audio
    >>> audio.start()

To stop recording::

    >>> audio.stop()

To play recording::

    >>> audio.play()

Supported Platforms
-------------------
Android

'''

from plyer.compat import string_types, text_type


class Audio(object):
    '''
    Audio facade.
    '''

    state = 'ready'
    _file_path = ''

    def __init__(self, file_path=None):
        super(Audio, self).__init__()
        self._file_path = file_path or self._file_path

    def start(self):
        '''
        Start record.
        '''
        self._start()
        self.state = 'recording'

    def stop(self):
        '''
        Stop record.
        '''
        self._stop()
        self.state = 'ready'

    def play(self):
        '''
        Play current recording.
        '''
        self._play()
        self.state = 'playing'

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, location):
        '''
        Location of the recording.
        '''
        assert isinstance(location, (string_types, text_type)), \
            'Location must be string or unicode'
        self._file_path = location

    # private

    def _start(self):
        raise NotImplementedError()

    def _stop(self):
        raise NotImplementedError()

    def _play(self):
        raise NotImplementedError()



#!/usr/bin/env python
'''
Setup.py
========

For MacOS or iOS install additional dependency PyOBJus::

    pip install https://github.com/kivy/pyobjus/zipball/master

For Android install additional dependency PyJNIus::

    pip install https://github.com/kivy/pyjnius/zipball/master
'''

from os.path import dirname, join
import plyer
import io

EXTRA_OPTIONS = {}

try:
    from setuptools import setup
    EXTRA_OPTIONS = dict(
        EXTRA_OPTIONS, **{
            'extras_require': {
                'ios': ['pyobjus'],
                'macosx': ['pyobjus'],
                'android': ['pyjnius'],
                'dev': ['mock', 'pycodestyle', 'pylint']
            }
        }
    )

except ImportError:
    from distutils.core import setup

CURDIR = dirname(__file__)
PACKAGES = [
    'plyer',
    'plyer.facades',
    'plyer.platforms',
    'plyer.platforms.linux',
    'plyer.platforms.android',
    'plyer.platforms.win',
    'plyer.platforms.win.libs',
    'plyer.platforms.ios',
    'plyer.platforms.macosx',
    'plyer.platforms.macosx.libs',
]

with io.open(join(CURDIR, "README.md"), encoding="utf8") as fd:
    README = fd.read()
with io.open(join(CURDIR, "CHANGELOG.md"), encoding="utf8") as fd:
    CHANGELOG = fd.read()

setup(
    name='plyer',
    version=plyer.__version__,
    description='Platform-independent wrapper for platform-dependent APIs',
    long_description=README + u"\n\n" + CHANGELOG + u"\n\n",
    long_description_content_type='text/markdown',
    author='Kivy team',
    author_email='mat@kivy.org',
    url='https://plyer.readthedocs.org/en/latest/',
    packages=PACKAGES,
    package_data={'': ['LICENSE', 'README.md']},
    package_dir={'plyer': 'plyer'},
    include_package_data=True,
    license=open(join(CURDIR, 'LICENSE')).read(),
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    **EXTRA_OPTIONS
)



'''
Module of Windows API helper for plyer.battery.
'''

__all__ = ('battery_status')


import ctypes
from plyer.platforms.win.libs import win_api_defs


def battery_status():
    '''
    Implementation of Windows system power status API for plyer.battery.
    '''

    status = win_api_defs.SYSTEM_POWER_STATUS()
    if not win_api_defs.GetSystemPowerStatus(ctypes.pointer(status)):
        raise Exception('Could not get system power status.')

    # pylint: disable=protected-access
    return dict(
        (field, getattr(status, field))
        for field, _ in status._fields_
    )



'''UniqueID facade.

Returns the following depending on the platform:

* **Android**: Android ID
* **OS X**: Serial number of the device
* **Linux**: Serial number using lshw
* **Windows**: MachineGUID from regkey
* **iOS**: UUID

Simple Example
--------------

To get the unique ID::

    >>> from plyer import uniqueid
    >>> uniqueid.id
    '1b1a7a4958e2a845'

.. versionadded:: 1.2.0

.. versionchanged:: 1.2.4
    On Android returns Android ID instead of IMEI.

Supported Platforms
-------------------
Android, iOS, Windows, OS X, Linux

'''


class UniqueID(object):
    '''
    UniqueID facade.
    '''

    @property
    def id(self):  # pylint: disable=invalid-name
        '''
        Property that returns the unique id of the platform.
        '''
        return self.get_uid()

    def get_uid(self):
        '''
        Public method for receiving unique ID via platform-specific
        API in plyer.platforms.
        '''
        return self._get_uid()

    # private

    def _get_uid(self):
        raise NotImplementedError()



'''
Camera
======

The :class:`Camera` is to capture pictures and make videos.

.. note::
        - On Android the `CAMERA` , `WRITE_EXTERNAL_STORAGE`,
          `READ_EXTERNAL_STORAGE` permissions are needed.

Simple Examples
---------------

Setup callback function.

    >>> from os.path import exists, join
    >>> from plyer import camera
    >>> def camera_callback(filepath):
    >>>     if(exists(filepath)):
    >>>         print "saved"
    >>>     else:
    >>>         print "unable to save."
    >>> filepath = 'path/to/your/file'
    >>> # e.g: filepath = join(App.get_running_app().user_data_dir, file_name)

To take picture::

    >>> file_name = "test.jpg"
    >>> camera.take_picture(filename=file_name,
    >>>                     on_complete=camera_callback)

Ta take a video::

    >>> file_name = "test.mp4"
    >>> camera.take_video(filename=file_name,
    >>>                   on_complete=camera_callback)

Supported Platforms
-------------------
Android, iOS

'''


class Camera(object):
    '''
    Camera facade.
    '''

    def take_picture(self, filename, on_complete):
        '''Ask the OS to capture a picture, and store it at filename.

        When the capture is done, on_complete will be called with the filename
        as an argument. If the callback returns True, the filename will be
        unlinked.

        :param filename: Name of the image file
        :param on_complete: Callback that will be called when the operation is
            done

        :type filename: str
        :type on_complete: callable
        '''
        self._take_picture(filename=filename, on_complete=on_complete)

    def take_video(self, filename, on_complete):
        '''Ask the OS to capture a video, and store it at filename.

        When the capture is done, on_complete will be called with the filename
        as an argument. If the callback returns True, the filename will be
        unlinked.

        :param filename: Name of the video file
        :param on_complete: Callback that will be called when the operation is
            done

        :type filename: str
        :type on_complete: callable
        '''
        self._take_video(filename=filename, on_complete=on_complete)

    # private

    def _take_picture(self, **kwargs):
        raise NotImplementedError()

    def _take_video(self, **kwargs):
        raise NotImplementedError()



'''
Android Storage Path
--------------------
'''

from os import listdir, access, R_OK
from os.path import join
from plyer.facades import StoragePath
from jnius import autoclass
from android import mActivity

Environment = autoclass('android.os.Environment')
Context = autoclass('android.content.Context')


class AndroidStoragePath(StoragePath):

    def _get_home_dir(self):
        return Environment.getDataDirectory().getAbsolutePath()

    def _get_external_storage_dir(self):
        return Environment.getExternalStorageDirectory().getAbsolutePath()

    def _get_sdcard_dir(self):
        '''
        .. versionadded:: 1.4.0
        '''
        # folder in /storage/ that is readable
        # and is not internal SD card
        path = None
        for folder in listdir('/storage'):
            folder = join('/storage', folder)
            if folder in self._get_external_storage_dir():
                continue
            if not access(folder, R_OK):
                continue
            path = folder
            break
        return path

    def _get_root_dir(self):
        return Environment.getRootDirectory().getAbsolutePath()

    def _get_documents_dir(self):
        return Environment.getExternalStoragePublicDirectory(
            Environment.DIRECTORY_DOCUMENTS).getAbsolutePath()

    def _get_downloads_dir(self):
        return Environment.getExternalStoragePublicDirectory(
            Environment.DIRECTORY_DOWNLOADS).getAbsolutePath()

    def _get_videos_dir(self):
        return Environment.getExternalStoragePublicDirectory(
            Environment.DIRECTORY_MOVIES).getAbsolutePath()

    def _get_music_dir(self):
        return Environment.getExternalStoragePublicDirectory(
            Environment.DIRECTORY_MUSIC).getAbsolutePath()

    def _get_pictures_dir(self):
        return Environment.getExternalStoragePublicDirectory(
            Environment.DIRECTORY_PICTURES).getAbsolutePath()

    def _get_application_dir(self):
        return mActivity.getFilesDir().getParentFile().getParent()


def instance():
    return AndroidStoragePath()



'''
Sms
====

The :class:`Sms` provides access to sending Sms from your device.

.. note::

    On Android your app needs the SEND_SMS permission in order to
    send sms messages.

.. versionadded:: 1.2.0

Simple Examples
---------------

To send sms::

    >>> from plyer import sms
    >>> recipient = 9999222299
    >>> message = 'This is an example.'
    >>> sms.send(recipient=recipient, message=message)

Supported Platforms
-------------------
Android, iOS

'''


class Sms(object):
    '''
    Sms facade.
    '''

    def send(self, recipient, message):
        '''
        Send SMS or open SMS interface.

        :param recipient: The reveiver
        :param message: the message

        :type recipient: number
        :type message: str
        '''
        self._send(recipient=recipient, message=message)

    # private

    def _send(self, **kwargs):
        raise NotImplementedError()



'''
Module of Android API for plyer.uniqueid.
'''

from jnius import autoclass
from plyer.platforms.android import activity
from plyer.facades import UniqueID

Secure = autoclass('android.provider.Settings$Secure')


class AndroidUniqueID(UniqueID):
    '''
    Implementation of Android uniqueid API.
    '''

    def _get_uid(self):
        return Secure.getString(
            activity.getContentResolver(),
            Secure.ANDROID_ID
        )


def instance():
    '''
    Instance for facade proxy.
    '''
    return AndroidUniqueID()



'''
Reference
Methods, Structures and Documentation adapted from.
https://msdn.microsoft.com/en-us/library/windows/desktop \
/ms705945%28v=vs.85%29.aspx
'''

from ctypes import *
from ctypes.wintypes import *
from sys import exit as sys_exit
from plyer.compat import PY2, xrange


def customresize(array, new_size):
    return (
        array._type_ * new_size
    ).from_address(
        addressof(array)
    )


wlanapi = windll.LoadLibrary('wlanapi.dll')


class GUID(Structure):
    _fields_ = [
        ('Data1', c_ulong),
        ('Data2', c_ushort),
        ('Data3', c_ushort),
        ('Data4', c_ubyte * 8),
    ]


# The WLAN_INTERFACE_STATE enumerated type indicates the state of an interface.
WLAN_INTERFACE_STATE = c_uint
(wlan_interface_state_not_ready,
 wlan_interface_state_connected,
 wlan_interface_state_ad_hoc_network_formed,
 wlan_interface_state_disconnecting,
 wlan_interface_state_disconnected,
 wlan_interface_state_associating,
 wlan_interface_state_discovering,
 wlan_interface_state_authenticating) = map(WLAN_INTERFACE_STATE,
                                            xrange(0, 8))


class WLAN_INTERFACE_INFO(Structure):
    '''
    The WLAN_INTERFACE_STATE enumerated type indicates the state of an
    interface.
    '''
    _fields_ = [
        ("InterfaceGuid", GUID),
        ("strInterfaceDescription", c_wchar * 256),
        ("isState", WLAN_INTERFACE_STATE)
    ]


class WLAN_INTERFACE_INFO_LIST(Structure):
    '''
    The WLAN_INTERFACE_INFO_LIST structure contains an array of NIC interface
    information.
    '''
    _fields_ = [
        ("NumberOfItems", DWORD),
        ("Index", DWORD),
        ("InterfaceInfo", WLAN_INTERFACE_INFO * 1)
    ]


DOT11_MAC_ADDRESS = c_ubyte * 6
WLAN_MAX_PHY_TYPE_NUMBER = 0x8
DOT11_SSID_MAX_LENGTH = 32
WLAN_REASON_CODE = DWORD

DOT11_BSS_TYPE = c_uint
(dot11_BSS_type_infrastructure,
 dot11_BSS_type_independent,
 dot11_BSS_type_any) = map(DOT11_BSS_TYPE, xrange(1, 4))

# The DOT11_PHY_TYPE enumeration defines an 802.11 PHY and media type.
DOT11_PHY_TYPE = c_uint
dot11_phy_type_unknown = 0
dot11_phy_type_any = 0
dot11_phy_type_fhss = 1
dot11_phy_type_dsss = 2
dot11_phy_type_irbaseband = 3
dot11_phy_type_ofdm = 4
dot11_phy_type_hrdsss = 5
dot11_phy_type_erp = 6
dot11_phy_type_ht = 7
dot11_phy_type_IHV_start = 0x80000000
dot11_phy_type_IHV_end = 0xffffffff

# The DOT11_AUTH_ALGORITHM enumerated type defines a wireless
# LAN authentication algorithm.
DOT11_AUTH_ALGORITHM = c_uint
DOT11_AUTH_ALGO_80211_OPEN = 1
DOT11_AUTH_ALGO_80211_SHARED_KEY = 2
DOT11_AUTH_ALGO_WPA = 3
DOT11_AUTH_ALGO_WPA_PSK = 4
DOT11_AUTH_ALGO_WPA_NONE = 5
DOT11_AUTH_ALGO_RSNA = 6
DOT11_AUTH_ALGO_RSNA_PSK = 7
DOT11_AUTH_ALGO_IHV_START = 0x80000000
DOT11_AUTH_ALGO_IHV_END = 0xffffffff

# The DOT11_CIPHER_ALGORITHM enumerated type defines a cipher
# algorithm for data encryption and decryption.
DOT11_CIPHER_ALGORITHM = c_uint
DOT11_CIPHER_ALGO_NONE = 0x00
DOT11_CIPHER_ALGO_WEP40 = 0x01
DOT11_CIPHER_ALGO_TKIP = 0x02
DOT11_CIPHER_ALGO_CCMP = 0x04
DOT11_CIPHER_ALGO_WEP104 = 0x05
DOT11_CIPHER_ALGO_WPA_USE_GROUP = 0x100
DOT11_CIPHER_ALGO_RSN_USE_GROUP = 0x100
DOT11_CIPHER_ALGO_WEP = 0x101
DOT11_CIPHER_ALGO_IHV_START = 0x80000000
DOT11_CIPHER_ALGO_IHV_END = 0xffffffff


class DOT11_SSID(Structure):
    '''
    A DOT11_SSID structure contains the SSID of an interface
    '''
    _fields_ = [
        ("SSIDLength", c_ulong),
        ("SSID", c_char * DOT11_SSID_MAX_LENGTH)
    ]


# Enumerated type to define the code of connection.
WLAN_CONNECTION_MODE = c_uint
(wlan_connection_mode_profile,
 wlan_connection_mode_temporary_profile,
 wlan_connection_mode_discovery_secure,
 wlan_connection_mode_discovery_unsecure,
 wlan_connection_mode_auto,
 wlan_connection_mode_invalid) = map(WLAN_CONNECTION_MODE, xrange(0, 6))


class NDIS_OBJECT_HEADER(Structure):
    '''
    This Structure packages the object type, version, and size information
    that is required in many NDIS (Netword Driver interface Specification)
    Structures.
    '''
    _fields_ = [
        ("Type", c_char),
        ("Revision", c_char),
        ("Size", c_ushort)]


class DOT11_BSSID_LIST(Structure):
    '''
    The DOT11_BSSID_LIST structure contains a list of basic service set (BSS)
    identifiers.
    '''
    _fields_ = [
        ("Header", NDIS_OBJECT_HEADER),
        ("uNumOfEntries", ULONG),
        ("uTotalNumOfEntries", ULONG),
        ("BSSIDs", DOT11_MAC_ADDRESS * 1)
    ]


class WLAN_CONNECTION_PARAMETERS(Structure):
    '''
    The WLAN_CONNECTION_PARAMETERS structure specifies the parameters used when
    using the WlanConnect function.
    '''
    _fields_ = [
        ("wlanConnectionMode", WLAN_CONNECTION_MODE),
        ("strProfile", LPCWSTR),
        ("pDot11Ssid", POINTER(DOT11_SSID)),
        ("pDesiredBssidList", POINTER(DOT11_BSSID_LIST)),
        ("dot11BssType", DOT11_BSS_TYPE),
        ("dwFlags", DWORD)]


# The `WlanConnect` attempts to connect to a specific network.
WlanConnect = wlanapi.WlanConnect
WlanConnect.argtypes = (HANDLE,
                        POINTER(GUID),
                        POINTER(WLAN_CONNECTION_PARAMETERS),
                        c_void_p)
WlanConnect.restype = DWORD

# The `WlanDisconnect` method disconnects an interface from its
# current network.
WlanDisconnect = wlanapi.WlanDisconnect
WlanDisconnect.argtypes = (HANDLE,
                           POINTER(GUID),
                           c_void_p)
WlanDisconnect.restype = DWORD

# Opens a connection to the server.
WlanOpenHandle = wlanapi.WlanOpenHandle
WlanOpenHandle.argtypes = (DWORD, c_void_p, POINTER(DWORD), POINTER(HANDLE))
WlanOpenHandle.restype = DWORD

# The WlanCloseHandle method closes the connection to the server.
WlanCloseHandle = wlanapi.WlanCloseHandle
WlanCloseHandle.argtypes = (HANDLE, c_void_p)
WlanCloseHandle.restype = DWORD


class WLAN_AVAILABLE_NETWORK(Structure):
    '''
    The WLAN_INTERFACE_INFO structure contains information about a wireless
    LAN interface.
    '''
    _fields_ = [
        ("ProfileName", c_wchar * 256),
        ("dot11Ssid", DOT11_SSID),
        ("dot11BssType", DOT11_BSS_TYPE),
        ("NumberOfBssids", c_ulong),
        ("NetworkConnectable", c_bool),
        ("wlanNotConnectableReason", WLAN_REASON_CODE),
        ("NumberOfPhyTypes", c_ulong),
        ("dot11PhyTypes", DOT11_PHY_TYPE * WLAN_MAX_PHY_TYPE_NUMBER),
        ("MorePhyTypes", c_bool),
        ("wlanSignalQuality", c_ulong),
        ("SecurityEnabled", c_bool),
        ("dot11DefaultAuthAlgorithm", DOT11_AUTH_ALGORITHM),
        ("dot11DefaultCipherAlgorithm", DOT11_CIPHER_ALGORITHM),
        ("Flags", DWORD),
        ("Reserved", DWORD)]


class WLAN_AVAILABLE_NETWORK_LIST(Structure):
    '''
    The WLAN_INTERFACE_INFO_LIST structure contains an array of NIC
    interface information.
    '''
    _fields_ = [
        ("NumberOfItems", DWORD),
        ("Index", DWORD),
        ("Network", WLAN_AVAILABLE_NETWORK * 1)]


# The WlanEnumInterfaces function enumerates all of the wireless LAN interfaces
# currently enabled on the local computer.
WlanEnumInterfaces = wlanapi.WlanEnumInterfaces
WlanEnumInterfaces.argtypes = (HANDLE,
                               c_void_p,
                               POINTER(POINTER(WLAN_INTERFACE_INFO_LIST)))
WlanEnumInterfaces.restype = DWORD

# The WlanGetAvailableNetworkList function retrieves the list of available
# networks on a wireless LAN interface.
WlanGetAvailableNetworkList = wlanapi.WlanGetAvailableNetworkList
WlanGetAvailableNetworkList.argtypes = (HANDLE,
                                        POINTER(GUID),
                                        DWORD,
                                        c_void_p,
                                        POINTER(POINTER(
                                            WLAN_AVAILABLE_NETWORK_LIST)))
WlanGetAvailableNetworkList.restype = DWORD

# The WlanFreeMemory function frees memory. Any memory returned from Native
# Wifi functions must be freed.
WlanFreeMemory = wlanapi.WlanFreeMemory
WlanFreeMemory.argtypes = [c_void_p]

wireless_interfaces = None
available = None
_dict = {}

# Private methods.


def _connect(network, parameters):
    '''
    Attempts to connect to a specific network.
    '''
    global _dict  # pylint: disable=global-statement
    wireless_interface = _dict[network]

    wcp = WLAN_CONNECTION_PARAMETERS()
    connection_mode = parameters['connection_mode']
    wcp.wlanConnectionMode = WLAN_CONNECTION_MODE(connection_mode)

    if connection_mode == 0 or connection_mode == 1:
        wcp.strProfile = LPCWSTR(connection_params["profile"])
    else:
        cnxp.strProfile = None

    dot11Ssid = DOT11_SSID()
    try:
        dot11Ssid.SSID = parameters["ssid"]
        dot11Ssid.SSIDLength = len(parameters["ssid"])
    except KeyError:
        dot11Ssid.SSID = network
        dot11Ssid.SSIDLength = len(network)
    wcp.pDot11Ssid = pointer(dot11Ssid)

    dot11bssid = DOT11_BSSID_LIST()
    bssid = parameters["bssidList"]
    dot11bssid.Header = bssid['Header']
    dot11bssid.uNumOfEntries = bssid['uNumOfEntries']
    dot11bssid.uTotalNumOfEntries = bssid['uTotalNumOfEntries']
    dot11bssid.BSSIDs = bssid['BSSIDs']

    wcp.pDesiredBssidList = pointer(bssidList)

    bssType = parameters["bssType"]
    wcp.dot11BssType = DOT11_BSS_TYPE(bssType)

    wcp.dwFlags = DWORD(parameters["flags"])

    NegotiatedVersion = DWORD()
    ClientHandle = HANDLE()

    wlan = WlanOpenHandle(1,
                          None,
                          byref(NegotiatedVersion),
                          byref(ClientHandle))
    if wlan:
        sys_exit(FormatError(wlan))
    pInterfaceList = pointer(WLAN_INTERFACE_INFO_LIST())
    wlan = WlanEnumInterfaces(ClientHandle, None, byref(pInterfaceList))
    if wlan:
        sys_exit(FormatError(wlan))

    try:
        wlan = WlanConnect(ClientHandle,
                           wireless_interface,
                           wcp,
                           None)
        if wlan:
            sys_exit(FormatError(wlan))
        WlanCloseHandle(ClientHandle, None)
    finally:
        WlanFreeMemory(pInterfaceList)


def _disconnect():
    '''
    To disconnect an interface form the current network.
    '''
    NegotiatedVersion = DWORD()
    ClientHandle = HANDLE()

    wlan = WlanOpenHandle(
        1,
        None,
        byref(NegotiatedVersion),
        byref(ClientHandle)
    )
    if wlan:
        sys_exit(FormatError(wlan))

    pInterfaceList = pointer(WLAN_INTERFACE_INFO_LIST())

    wlan = WlanEnumInterfaces(ClientHandle, None, byref(pInterfaceList))
    if wlan:
        sys_exit(FormatError(wlan))

    result = None
    try:
        ifaces = customresize(
            pInterfaceList.contents.InterfaceInfo,
            pInterfaceList.contents.NumberOfItems
        )

        # find each available network for each interface
        for iface in ifaces:
            wlan = WlanDisconnect(
                ClientHandle,
                byref(iface.InterfaceGuid),
                None
            )
            if wlan:
                sys_exit(FormatError(wlan))
            WlanCloseHandle(ClientHandle, None)

    finally:
        WlanFreeMemory(pInterfaceList)
        result = get_available_wifi()

    return result


def _start_scanning():
    '''
    Private method for scanning and returns the available devices.
    '''
    global available  # pylint: disable=global-statement
    global wireless_interfaces  # pylint: disable=global-statement
    NegotiatedVersion = DWORD()
    ClientHandle = HANDLE()

    wlan = WlanOpenHandle(
        1,
        None,
        byref(NegotiatedVersion),
        byref(ClientHandle)
    )

    if wlan:
        sys_exit(FormatError(wlan))

    # find all wireless network interfaces
    pInterfaceList = pointer(WLAN_INTERFACE_INFO_LIST())
    wlan = WlanEnumInterfaces(ClientHandle, None, byref(pInterfaceList))
    if wlan:
        sys_exit(FormatError(wlan))

    result = None
    try:
        ifaces = customresize(
            pInterfaceList.contents.InterfaceInfo,
            pInterfaceList.contents.NumberOfItems
        )

        # find each available network for each interface
        wireless_interfaces = ifaces
        for iface in ifaces:
            pAvailableNetworkList = pointer(WLAN_AVAILABLE_NETWORK_LIST())
            wlan = WlanGetAvailableNetworkList(
                ClientHandle,
                byref(iface.InterfaceGuid),
                0,
                None,
                byref(pAvailableNetworkList)
            )

            if wlan:
                sys_exit(FormatError(wlan))

            try:
                avail_net_list = pAvailableNetworkList.contents
                networks = customresize(
                    avail_net_list.Network,
                    avail_net_list.NumberOfItems
                )

                # Assigning the value of networks to the global variable
                # `available`, so it could be used in other methods.
                available = networks
                _make_dict()
                wlan = WlanDisconnect(
                    ClientHandle,
                    byref(iface.InterfaceGuid),
                    None
                )

                if wlan:
                    sys_exit(FormatError(wlan))
                WlanCloseHandle(ClientHandle, None)

            finally:
                WlanFreeMemory(pAvailableNetworkList)

    finally:
        WlanFreeMemory(pInterfaceList)
        result = get_available_wifi()

    return result


def _get_network_info(name):
    '''
    returns the list of the network selected in a dict.
    '''
    global available  # pylint: disable=global-statement
    global _dict  # pylint: disable=global-statement

    net = _dict[name]
    dot11BssType = net.dot11BssType
    dot11DefaultAuthAlgorithm = net.dot11DefaultAuthAlgorithm
    dot11DefaultCipherAlgorithm = net.dot11DefaultCipherAlgorithm
    dot11PhyTypes = net.dot11PhyTypes[0]
    dot11Ssid = net.dot11Ssid
    wlanNotConnectableReason = net.wlanNotConnectableReason
    wlanSignalQuality = net.wlanSignalQuality
    return {"dot11BssType": dot11BssType,
            "dot11DefaultAuthAlgorithm": dot11DefaultAuthAlgorithm,
            "dot11DefaultCipherAlgorithm": dot11DefaultCipherAlgorithm,
            "dot11PhyTypes": dot11PhyTypes,
            "SSID": dot11Ssid.SSID,
            "SSIDLength": dot11Ssid.SSIDLength,
            "wlanNotConnectableReason": wlanNotConnectableReason,
            "wlanSignalQuality": wlanSignalQuality}


def _make_dict():
    '''
    Prepares a dict so it could store network information.
    '''
    global available  # pylint: disable=global-statement
    global _dict  # pylint: disable=global-statement
    _dict = {}
    for network in available:
        # if bytes, dict['name'] throws an error on py3 if not b'name'
        if PY2:
            _dict[unicode(network.dot11Ssid.SSID)] = network
        else:
            _dict[network.dot11Ssid.SSID.decode('utf-8')] = network


def _get_available_wifi():
    '''
    returns the available wifi networks.
    '''
    global _dict  # pylint: disable=global-statement
    return _dict


def _is_enabled():
    '''
    Reason for returning true is explained in widi facade.
    /plyer/facades/wifi.py
    '''
    return True

# public methods.


def is_enabled():
    '''
    calls private method `_is_enabled` and returns the result.
    '''
    return _is_enabled()


def connect(network, parameters):
    '''
    Connect to a network with following parameters.
    '''
    _connect(network=network, parameters=parameters)


def disconnect():
    '''
    Disconnect from a network.
    '''
    _disconnect()


def start_scanning():
    '''
    Start scanning for available wifi networks available.
    '''
    return _start_scanning()


def get_network_info(name):
    '''
    return the wifi network info.
    '''
    return _get_network_info(name=name)


def get_available_wifi():
    '''
    return the available wifi networks available
    '''
    return _get_available_wifi()



'''
Android gravity
---------------------
'''

from jnius import autoclass
from jnius import cast
from jnius import java_method
from jnius import PythonJavaClass

from plyer.facades import Gravity
from plyer.platforms.android import activity

Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class GravitySensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(GravitySensorListener, self).__init__()

        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_GRAVITY
        )

        self.values = [None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self,
            self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:3]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidGravity(Gravity):

    def __init__(self):
        super(AndroidGravity, self).__init__()
        self.state = False

    def _enable(self):
        if not self.state:
            self.listener = GravitySensorListener()
            self.listener.enable()
            self.state = True

    def _disable(self):
        if self.state:
            self.state = False
            self.listener.disable()
            del self.listener

    def _get_gravity(self):
        if self.state:
            return tuple(self.listener.values)
        else:
            return (None, None, None)

    def __del__(self):
        if self.state:
            self._disable()
        super(self.__class__, self).__del__()


def instance():
    return AndroidGravity()



from jnius import autoclass
from jnius import cast
from jnius import java_method
from jnius import PythonJavaClass

from plyer.facades import Light
from plyer.platforms.android import activity

Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class LightSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(LightSensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)
        self.sensor = self.SensorManager.getDefaultSensor(Sensor.TYPE_LIGHT)
        self.value = None

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.value = event.values[0]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidLight(Light):

    listener = None

    def _get_illumination(self):
        if self.listener and self.listener.value:
            light = self.listener.value
            return light

    def _enable(self):
        if not self.listener:
            self.listener = LightSensorListener()
            self.listener.enable()

    def _disable(self):
        if self.listener:
            self.listener.disable()
            delattr(self, 'listener')


def instance():
    return AndroidLight()
